"""
"""
from abc import ABC, abstractmethod
from collections import OrderedDict
import enum
import importlib
from pkg_resources import iter_entry_points
import sys

# ProgressMeter statuses
class Status(enum.Enum):
    UNSTARTED = 0
    STARTED = 1
    FINISHED = 2


class ProgressMeterError(Exception):
    pass


class PluginManager:
    def __init__(self):
        self.plugins = None

    def set_plugin_options(self, config=None, **kwargs):
        pass

    def load_plugins(self, names=None, exclusive=False):
        plugins = {}
        for entry_point in iter_entry_points(group='pokrok'):
            plugin = entry_point.load()
            if plugin.installed:
                plugins[entry_point.name] = plugin

        if names:
            ordered_plugins = OrderedDict()
            for name in names:
                if name in plugins:
                    ordered_plugins[name] = plugins[name]
            if not exclusive:
                for name in (set(plugins.keys()) - set(ordered_plugins.keys())):
                    ordered_plugins[name] = plugins[name]
            self.plugins = ordered_plugins
        else:
            self.plugins = plugins

    def has_plugin(self, name):
        if not self.plugins:
            self.load_plugins()
        return name in self.plugins

    def get_plugin(self, name):
        if not self.has_plugin(name):
            raise ValueError("No such plugin: {}".format(name))
        if isinstance(self.plugins[name], type):
            self.plugins[name] = self.plugins[name]()
        return self.plugins[name]

    def get_first_plugin(self, sized=None, widgets=None):
        """Returns the first plugin that can provide a progress meter of the
        specified configuration.

        Args:
            sized: Whether a sized ProgressMeter is required.
            widgets: The desired style.

        Returns:
            A ProgressMeter instance if one is found that can provide the
            specified configuration, else None.
        """
        if self.plugins is None:
            self.load_plugins()

        if len(self.plugins) == 0:
            return None

        plugin_names = tuple(self.plugins.keys())
        if sized is None and widgets is None:
            return self.get_plugin(plugin_names[0])

        for plugin_name in plugin_names:
            plugin = self.get_plugin(plugin_name)
            if plugin.provides(sized, widgets):
                return plugin

        return None


class ProgressMeterFactory(ABC):
    """Plugin base class. A plugin's entry point must provide an instance of
    a ProgressMeterFactory.
    """
    @property
    @abstractmethod
    def name(self):
        """The plugin's name.
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def installed(self):
        """Import the package(s) depended upon by this module.

        Returns:
            True if all required packages are successfully imported.
        """
        raise NotImplementedError()

    @abstractmethod
    def provides(self, sized, widgets=None):
        """Whether the plugin can provide a progress meter that conforms to the
        specified style. The plugin is free to interpret this request as it sees
        fit, including ignoring or substituting widgets.

        Args:
            sized: Boolean, whether the user is requesting a sized progress
                meter.
            widgets: The Style requested, or None to use the plugin's default
                style.

        Returns:
            True if the plugin can provide a conforming progress meter.
        """
        raise NotImplementedError()

    @abstractmethod
    def create(self, size=None, widgets=None, desc=None, start=None, **kwargs):
        """Create a progress meter that conforms to the requested style.

        Args:
            size: Size of the progress meter.
            widgets: The desired style, or None to use the plugin's default
                style.
            desc: String description to print next to the progress bar.
            start: Counter start value.
            kwargs: Additional keyword arguments to pass to the underlying
                progress meter package. The plugin is free to interpret these
                arguments as it wishes, including ignoring them.

        Returns:
            A ProgressMeter instance.
        """
        raise NotImplementedError()

    @abstractmethod
    def iterate(
            self, iterable, size=None, widgets=None, desc=None, start=None,
            **kwargs):
        """Wrap an iterable with a progress meter.

        Args:
            iterable: The iterable to wrap.
            size: Size of the progress meter.
            widgets: The desired style, or None to use the plugin's default
                style.
            desc: String description to print next to the progress bar.
            start: Counter start value.
            kwargs: Additional keyword arguments to pass to the underlying
                progress meter package. The plugin is free to interpret these
                arguments as it wishes, including ignoring them.

        Returns:
            An iterable. Returns the original iterable if there were any
            problems creating the wrapper.

        """
        raise NotImplementedError()


class DefaultProgressMeterFactory(ProgressMeterFactory):
    """Default implementation of ProgressMeterFactory. Assumes that a) the
    underlying package is in a module that can be imported using importlib
    (if it is installed), b) the package supports any desired progress
    meter configuration (override `provides` if this isn't true), and c)
    the ProgressMeter subclass has a constructor with signature
    (module, size, widgets, desc, start, **kwargs).

    Args:
        name: The plugin name. Also used as the module name if `module_name`
            is None.
        progress_meter_class: The ProgressMeter subclass to instantiate.
        style_superset: A Style object that lists all of the widgets supported
            by the plugin.
        module_name: The name of the module, if different from `name`.
    """
    def __init__(
            self, name, progress_meter_class, style_superset,
            module_name=None):
        super().__init__()
        self._name = name
        self._package = module_name or name
        self._module = None
        self._progress_meter_class = progress_meter_class
        self._style_superset = style_superset

    @property
    def name(self):
        return self._name

    def installed(self):
        return self._load_module()

    def provides(self, sized, widgets=None):
        if not (widgets and self._style_superset):
            return True

        provided_widgets = self._style_superset.get_widgets(sized)
        if not provided_widgets:
            return False

        return len(set(widgets) - set(provided_widgets)) == 0

    def create(self, size=None, widgets=None, desc=None, start=None, **kwargs):
        self._load_module()
        return self._progress_meter_class(
            mod=self._module, size=None, widgets=None, desc=None, start=None,
            **kwargs)

    def iterate(
            self, iterable, size=None, widgets=None, desc=None, start=None,
            **kwargs):
        pbar = self.create(size, widgets, desc, start, **kwargs)
        if pbar:
            with pbar:
                for item in iterable:
                    yield item
                    pbar.increment()
        else:
            return iterable

    def _load_module(self):
        if self._module is None:
            try:
                self._module = importlib.import_module(self._package)
            except ImportError:
                self._module = False
        return self._module not in (None, False)


class ProgressMeter(ABC):
    """ProgressMeter interface to be implemented by the plugin.
    """
    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()

    @property
    @abstractmethod
    def is_sized(self):
        raise NotImplementedError()

    @property
    @abstractmethod
    def status(self):
        """The current status of the ProgressMeter. Must be one of the
        enumerated values.
        """
        raise NotImplementedError()

    def start(self):
        """Initialize and show the progress meter.
        """
        pass

    def finish(self):
        """Signal that the task is complete. The progress meter should update
        its display accordingly (if applicable).
        """
        pass

    @abstractmethod
    def increment(self, n=1):
        """Increment the progress meter by a fixed amount (default=1).

        Args:
            n: The amount to increment.

        Returns:
            The current value after incrementing.
        """
        raise NotImplementedError()

    def _check_status(self, status=Status.STARTED, error=True):
        """Checks that the ProgressMeter's status matches `stats`.

        Args:
            status: The status to check.
            error: If True, an error is raised if the ProgressMeter does not
                have the specified status.

        Returns:
            True if the ProgressMeter is has the specified status. False if the
            ProgressMeter does not have the specified status and `error` is
            False.

        Raises:
            ProgressMeterError if the ProgressMeter does not have the
            specified status and `error` is True.
        """
        if self.status != status:
            if error:
                raise ProgressMeterError(
                    "The ProgressMeter status {} differs from expected status "
                    "{}".format(self.status, status))
            else:
                return False

        return True


class BaseProgressMeter(ProgressMeter):
    """Default implementation of ProgressMeter. Subclasses only need to
    implement `increment()`.
    """
    def __init__(self, size=None):
        self._status = Status.UNSTARTED
        self.size = size

    @property
    def is_sized(self):
        return self.size is not None

    @property
    def status(self):
        return self._status

    def start(self):
        self._check_status(Status.UNSTARTED)
        self._status = Status.STARTED

    def finish(self):
        self._check_status(Status.STARTED)
        self._status = Status.FINISHED


class SimpleTextProgressMeter(BaseProgressMeter):
    """Simplest progress meter implementation. Simply prints a character each
    time it is incremented.
    """
    def __init__(self, char='*', file=sys.stderr, **kwargs):
        super().__init__()
        self.char = char
        self.file = file

    def increment(self, n=1):
        print(self.char * n, file=self.file)



"""Main pokrok API.

All functions support the following basic keyword arguments. Additional keyword
arguments can be passed to whichever plugin is selected, and that plugin can
choose whether or not to use those arguments.

* start: The starting value of the progress counter. Only value if `size` is
  also specified.
* size: The total size (i.e. max counter value) of the progress bar.
* style: The desired style; a string name of a pre-defined style, or a Style
  object.
* desc: A string description to display next to the progress bar.
* plugin_name: Name of a specific plugin to use.

"""
from collections.abc import Sized
import json
import math
import os

import pokrok.plugins
import pokrok.styles
from pokrok.styles import Style, Widget

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


class ProgressFactory:
    def __init__(self):
        self.plugins = pokrok.plugins.PluginManager()
        self.styles = pokrok.styles.StyleManager()
        self.configured = False

    @property
    def default_paths(self):
        return [
            os.path.join(path, 'pokrok.json')
            for path in (
                os.getcwd(),
                os.path.abspath(os.path.expanduser('~')))
        ]

    def configure(
            self, filename=None, plugin_names=None, exclusive=False,
            styles=None, **kwargs):
        if not (filename or self.configured):
            for fname in self.default_paths:
                if os.path.exists(fname):
                    filename = fname
                    break

        if filename:
            if os.path.exists(filename):
                with open(filename, 'rt') as inp:
                    config = json.load(inp)
            else:
                config = json.load(stream_from_package(filename))
                if config is None:
                    raise ValueError("File not found: {}".format(filename))
            self.plugins.set_plugin_options(config)
            self.styles.set_style_options(config)
            self.configured = True

        if plugin_names:
            self.plugins.load_plugins(plugin_names, exclusive)
        if kwargs:
            self.plugins.set_plugin_options(**kwargs)
        if styles:
            self.styles.update(**styles)

    def create(
            self, iterable=None, size=None, style='default', plugin_name=None,
            **kwargs):
        """Create a progress meter. All parameters are optional. The default
        behavior (i.e. when just calling `create()`) is to return an unsized
        ProgressMeter with default style.

        Args:
            iterable: An iterable to wrap.
            size: The size of the progress meter, or None for an unsized
                progress meter.
            style: The desired style.
            plugin_name: The name of a specific plugin to use.

        Returns:
            A ProgressMeter if `iterable` is None, otherwise an iterable, or
            None if a ProgressMeter could not be created.

        Raises:
            ProgressMeterError if a specific plugin is requested and is not
            available or does not support the requested configuration.
        """
        if not self.configured:
            self.configure()

        if isinstance(style, str):
            style = self.styles[style] if style else None

        sized = size is not None
        widgets = style.get_widgets(sized)

        if plugin_name:
            if not self.plugins.has_plugin(plugin_name):
                raise ValueError(
                    "Plugin {} is not supported".format(plugin_name))
            plugin = self.plugins.get_plugin(plugin_name)
            if not plugin.provides(sized, widgets):
                raise ValueError(
                    "Plugin {} does not support the requested configuration "
                    "(sized={}, style={})".format(plugin_name, sized, widgets))
        else:
            plugin = self.plugins.get_first_plugin(sized, widgets)

        if plugin and iterable:
            return plugin.iterate(
                iterable, size=size, widgets=widgets, **kwargs)
        elif plugin:
            return plugin.create(size=size, widgets=widgets, **kwargs)
        elif iterable:
            return iterable
        else:
            return None


def stream_from_package(spec):
    try:
        package, path = spec.split(':')
        import pkg_resources as pr
        if pr.resource_exists(package, path):
            return pr.resource_stream(package, path)
    except:
        pass
    return None


# Singleton factory class
FACTORY = ProgressFactory()


def set_plugins(names, exclusive=False):
    """High-level configuration of progress meter packages.

    Args:
        names: Names of plugins to prefer, in order.
        exclusive: Whether the listed packages should be the only ones allowed.
    """
    FACTORY.configure(plugin_names=names, exclusive=exclusive)


def set_styles(**styles):
    """Convenience method for configuring progress meter styles.

    Args:
        styles: keyword arguments, where the name is the style name and the
            value is a pokrok.styles.Style object.
    """
    FACTORY.configure(styles=styles)


def configure(**kwargs):
    """Low-level configuration. This is just a pass-through to
    FACTORY.configure().

    Args:
        kwargs: Keyword arguments to pass to FACTORY.configure().
    """
    FACTORY.configure(**kwargs)


def progress_range(start, stop=None, step=1, **kwargs):
    """Iterate over a range while showing a progress bar.

    Args:
        start: Range start.
        stop: Range stop.
        step: Range step.
        kwargs: Additional arguments - see package documentation.

    Returns:
        An iterable.
    """
    if stop is not None:
        r = range(start, stop, step)
        size = math.ceil((stop - start) / step)
    else:
        r = range(start)
        size = start
    return progress_iter(r, size=size, **kwargs)


def progress_file(filename, mode, **kwargs):
    """Iterate over a file while showing a progress bar.

    Args:
        filename: The name of the file to open and iterate over.
        mode: The file mode (must be readable).
        kwargs: Additional arguments - see package documentation.

    Yields:
        Lines from the file.
    """
    with open(filename, mode) as f:
        yield from progress_iter(f, **kwargs)


def progress_iter(iterable, size = None, **kwargs):
    """Wrap an iterable in a progress bar.
    
    Args:
        iterable: The iterable to wrap.
        size: The number of items that will be iterated over by the iterable.
            If None and this iterable happens to be Sized, the size will be
            determined using `len`.
        kwargs: Additional arguments - see package documentation.

    Returns:
        An iterable.
    """
    if size is None and isinstance(iterable, Sized):
        size = len(iterable)
    return FACTORY.create(iterable=iterable, size=size, **kwargs)


def progress_meter(**kwargs):
    """
    Create a progress meter.

    Args:
        kwargs: see package documentation.

    Returns:
        A ProgressMeter.
    """
    return FACTORY.create(**kwargs)



from pokrok.plugins import (
    DefaultProgressMeterFactory, BaseProgressMeter, Status)
from pokrok.styles import Style, Widget


class TqdmProgressMeterFactory(DefaultProgressMeterFactory):
    def __init__(self):
        style_superset = Style([
            Widget.BAR, Widget.ETA, Widget.ELAPSED, Widget.SPINNER
        ])
        super().__init__('tqdm', TqdmProgressMeter, style_superset)

    def iterate(
            self, iterable, size=None, widgets=None, desc=None, start=None,
            **kwargs):
        if self._load_module():
            return self._module.tqdm(
                iterable, total=size, desc=desc, initial=start or 0, **kwargs)
        else:
            return iterable


class TqdmProgressMeter(BaseProgressMeter):
    def __init__(self, mod, size, widgets, desc, start, **kwargs):
        super().__init__(size)
        self.tqdm = mod.tqdm(
            total=size, desc=desc, initial=start or 0, **kwargs)

    def finish(self):
        super().finish()
        self.tqdm.close()

    def increment(self, n=1):
        self._check_status(Status.STARTED)
        self.tqdm.update(n)



from pokrok.plugins import (
    DefaultProgressMeterFactory, BaseProgressMeter, Status)
from pokrok.styles import Style, Widget


class HaloProgressMeterFactory(DefaultProgressMeterFactory):
    def __init__(self):
        style_superset = Style(unsized=[Widget.SPINNER])
        super().__init__('halo', HaloProgressMeter, style_superset)

    def iterate(
            self, iterable, size=None, widgets=None, desc=None, start=None,
            **kwargs):
        if self._load_module():
            with self._module.Halo(text=desc or '', **kwargs):
                yield from iterable
        else:
            return iterable

    def provides(self, sized, widgets=None):
        if sized:
            return False
        return not widgets or (len(widgets) == 1 and widgets[1] == Widget.SPINNER)


class HaloProgressMeter(BaseProgressMeter):
    def __init__(self, mod, size, widgets, desc, start, **kwargs):
        super().__init__(size)
        self.spinner = mod.Halo(text=desc or '', **kwargs)

    def start(self):
        super().start()
        self.spinner.start()

    def finish(self):
        super().finish()
        self.spinner.succeed()

    def increment(self, n=1):
        pass



"""Build pokrok.
"""
import os.path
import sys

from setuptools import setup, find_packages
import versioneer

if sys.version_info < (3, 4):
    sys.stdout.write("At least Python 3.4 is required.\n")
    sys.exit(1)

setup(
    name = 'pokrok',
    version = versioneer.get_version(),
    cmdclass = versioneer.get_cmdclass(),
    author = 'John Didion',
    author_email = 'johndidion@gmail.com',
    url = 'https://github.com/jdidion/pokrok',
    description = 'Simple API for progress bars using any of several supported libraries',
    license = 'MIT',
    packages = find_packages(),
    tests_require = ['pytest'], #, 'jinja2', 'pysam'],
    entry_points = {
        'pokrok' : [
            'tqdm=pokrok.plugins.tqdm:TqdmProgressMeterFactory',
            'halo=pokrok.plugins.halo:HaloProgressMeterFactory'
        ]
    },
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: User Interfaces",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6"
    ]
)



import enum


class Widget(enum.Enum):
    BAR = 'BAR'
    ETA = 'ETA'
    ELAPSED = 'ELAPSED'
    SPINNER = 'SPINNER'
    COUNTER = 'COUNTER'
    PERCENT = 'PERCENT'


class StyleManager(dict):
    def __init__(self, default_style=None):
        super().__init__()
        self['default'] = default_style or Style()

    def set_style_options(self, config, **kwargs):
        if 'styles' in config:
            for name, style_config in config['styles'].items():
                self[name] = Style(**style_config)
        if 'default_style' in config:
            if isinstance(config['default_style'], str):
                self['default'] = self[config['default_style']]
            else:
                self['default'] = Style(**config['default_style'])


class Style:
    """
    Args:
        widgets: Widgets to use for `sized` and `unsized` unless they are not
            None.
        sized: Widgets for sized progress meters.
        unsized: Widgets specifically for unsized progress meters.s
    """
    def __init__(self, widgets=None, sized=None, unsized=None):
        if widgets:
            self.sized = self._resolve_widgets(sized or widgets)
            self.unsized = self._resolve_widgets(unsized or widgets)
        else:
            self.sized = self._resolve_widgets(sized)
            self.unsized = self._resolve_widgets(unsized)

    def _resolve_widgets(self, widgets):
        return [
            Widget[w] if isinstance(w, str) else w
            for w in widgets
        ]

    def get_widgets(self, sized):
        return self.sized if sized else self.unsized
